from __future__ import annotations

import asyncio
import logging
from typing import List, Callable, Optional, Awaitable

from google.protobuf import empty_pb2

from tq42.exception_handling import handle_generic_sdk_errors
from com.terraquantum.channel.v1alpha1.create_channel_pb2 import CreateChannelResponse

# important for re-export
from com.terraquantum.channel.v1alpha1.channel_message_pb2 import (
    ChannelMessage,
    Ask,
    Parameter,  # pylint:disable=unused-import # noqa: F401
    Tell,
    DataAcknowledge,
)

from typing import TYPE_CHECKING

# only import the stuff for type hints -> avoid circular imports
if TYPE_CHECKING:
    from tq42.client import TQ42Client


from tq42.utils.timers import AsyncTimedIterable


class Channel:
    """
    Class to create and connect to a channel
    """

    id: str
    client: TQ42Client

    def __init__(self, client: TQ42Client, id: str):
        self.client = client
        self.id = id

    def __repr__(self) -> str:
        return f"<Channel Id={self.id}>"

    def __str__(self) -> str:
        return self.id

    @staticmethod
    @handle_generic_sdk_errors
    async def create(client: TQ42Client) -> Channel:
        empty = empty_pb2.Empty()
        res: CreateChannelResponse = await client.channel_client.CreateChannel(
            request=empty, metadata=client.metadata
        )
        return Channel(client=client, id=res.channel_id)

    async def connect(
        self,
        callback: Callable[[Ask], Awaitable[Tell]],
        finish_callback: Callable,
        max_duration_in_sec: Optional[int] = None,
        message_timeout_in_sec: int = 30,
    ) -> None:
        """
        Connects to the stream and handles every message with the provided callback to create an answer.
        ASK gets into the callback and then we expect a TELL answer

        :param callback: Async callback that handles an ASK message and returns a TELL message
        :param finish_callback: Callback that is called when we finish the connection
        :param int max_duration_in_sec: Timeout for whole connection in seconds. `None` -> no timeout for overall flow
        :param int message_timeout_in_sec: Timeout between messages in seconds. Main way to end the connection.
        """

        async def _handle():
            try:
                timed_stream = AsyncTimedIterable(call, timeout=message_timeout_in_sec)
                async for incoming in timed_stream:
                    logging.debug(f"User received {incoming=}")
                    if incoming.HasField("ask_data"):
                        ask_msg: ChannelMessage = incoming
                        ack_msg = ChannelMessage(
                            acknowledge_data=DataAcknowledge(
                                id=incoming.sequential_message_id
                            )
                        )
                        tell = await callback(incoming.ask_data)
                        await call.write(ack_msg)
                        logging.debug(
                            f"User Sent ack {incoming.sequential_message_id=}"
                        )

                        tell_msg = ChannelMessage(
                            sequential_message_id=(ask_msg.sequential_message_id + 1),
                            tell_data=tell,
                        )
                        await call.write(tell_msg)

            except asyncio.CancelledError:
                logging.debug("Stream reading was cancelled.")
            except asyncio.TimeoutError:
                logging.debug("Stream finished because of the provided timeouts")
            except Exception as e:
                logging.debug(f"An unexpected error occurred in the stream: {e}")

        metadata: tuple = (
            *self.client.metadata,
            ("channel-id", self.id),
        )
        call = self.client.channel_client.ConnectChannelCustomer(metadata=metadata)

        await call.write(ChannelMessage())
        await asyncio.wait_for(_handle(), timeout=max_duration_in_sec)
        await call.done_writing()

        finish_callback()


@handle_generic_sdk_errors
def list_all(client: TQ42Client) -> List[Channel]:
    """
    List all channels for a given user.

    For details, see (TODO: update link once a new documentation URL is created)
    """
    raise NotImplementedError("This is a functionality still to come")



import importlib
import json
from typing import Union
import keyring

from com.terraquantum.experiment.v3alpha1.experimentrun import (
    create_experiment_run_request_pb2 as create_exp_run,
)
from com.terraquantum.experiment.v1.experimentrun.experiment_run_pb2 import (
    HardwareProto,
)
from com.terraquantum.experiment.v1.experimentrun.algorithm import shared_pb2

import tq42.utils.dirs as dirs
from tq42.utils import file_handling
from tq42.exceptions import NoMatchingAttributeError
from keyring.errors import NoKeyringError, InitError, PasswordSetError, KeyringLocked


def get_id(input):
    input = str(input)
    inputs = input.strip().splitlines()
    for input in inputs:
        pair_result = input.split(":")
        if len(pair_result) > 1 and pair_result[0] == "id":
            return pair_result[1].strip()

    return ""


def get_hw_configurations():
    hardware_configs_file_path = dirs.full_path(
        dirs.text_files_dir(), "hardware_configs.json"
    )
    with open(hardware_configs_file_path, "r") as json_file:
        loaded_data = json.load(json_file)
        return loaded_data


def get_hardware_num(name: Union[int, str]) -> int:
    try:
        if isinstance(name, str):
            return getattr(HardwareProto, name.upper())
        if isinstance(name, int):
            # TODO: we need to validate this is in range of the numbers we have in the enums
            return name
    except AttributeError:
        raise NoMatchingAttributeError("There is no hardware type {}".format(name))


def get_algo_num(name: Union[int, str]) -> int:
    try:
        if isinstance(name, str):
            return getattr(shared_pb2.AlgorithmProto, name.upper())
        if isinstance(name, int):
            # TODO: we need to validate this is in range of the numbers we have in the enums
            return name
    except AttributeError:
        raise NoMatchingAttributeError("There is no algorithm type {}".format(name))


def find_oneof_field_name(msg_type: str) -> str:
    """
    For the ExperimentRun we need to find out which field name matches our message type.
    This function looks for the matching oneof field in the CreateExperimentRunRequest protobuf message.
    """
    for oneof_field in create_exp_run.CreateExperimentRunRequest.DESCRIPTOR.oneofs:
        if oneof_field.name == "metadata":
            for field in oneof_field.fields:
                if field.message_type.name == msg_type:
                    return field.name


def dynamic_create_exp_run_request(
    parameters: dict, algo: int, exp_id: str, hardware: int
) -> create_exp_run.CreateExperimentRunRequest:
    """
    Dynamically creates a CreateExpRunRequest using the mapping provided as custom options of the AlgorithmProto enum.
    Tries dynamic import of this type and parses the parameters.
    """
    in_type = (
        shared_pb2.AlgorithmProto.DESCRIPTOR.values_by_number[algo]
        .GetOptions()
        .Extensions[shared_pb2.in_type]
    )
    path, msg_type = in_type.split(":")
    path = path.replace(".proto", "_pb2")
    path = path.replace("/", ".")
    module = importlib.import_module(path)

    # try parsing locally to reduce errors when communicating with the API
    try:
        getattr(module, msg_type)(**parameters)
    except Exception as e:
        raise NoMatchingAttributeError(str(e))

    # if parsing worked pass the params with the corresponding field name dynamically as keyword arguments
    field_name = find_oneof_field_name(msg_type=msg_type)
    keyword_args = {field_name: parameters}

    experiment_request = create_exp_run.CreateExperimentRunRequest(
        algorithm=algo,
        experiment_id=exp_id,
        hardware=hardware,
        **keyword_args,
    )
    return experiment_request


def save_token(service_name: str, backup_save_path: str, token: str) -> str:
    try:
        save_location = "keyring"
        keyring.set_password(
            service_name=service_name,
            username="username",
            password=token,
        )
        return save_location

    except (NoKeyringError, InitError, PasswordSetError):
        file_handling.write_to_file(backup_save_path, token)
        return backup_save_path


def get_token(service_name: str, backup_save_path: str) -> str:
    try:
        return keyring.get_password(
            service_name=service_name,
            username="username",
        )

    except (NoKeyringError, InitError, KeyringLocked):
        return file_handling.read_file(backup_save_path)



import unittest

from tq42.project import Project
from tq42.utils.utils_for_cache import clear_cache
from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliExperimentGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_exp_list(self):
        args = parse_args(["exp", "list", "--proj=" + self.proj])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

        # need a default set for just running exp list
        Project(client=self.get_client(), id=self.proj).set()
        args = parse_args(["exp", "list"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)
        # clear the defaults to have each test being individual
        clear_cache()

    def test_exp_set_friendly_name(self):
        args = parse_args(
            ["exp", "set-friendly-name", "FRIENDLY_NAME", "--exp=" + self.exp]
        )
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)



import unittest

from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalInvokeCliProjectGroup(unittest.TestCase, FunctionalCLITestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_proj_show(self):
        args = parse_args(["proj", "show"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_proj_list(self):
        args = parse_args(["proj", "list"])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

        args = parse_args(["proj", "list", "--org=" + self.org])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_proj_set(self):
        args = parse_args(["proj", "set", self.proj])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_proj_set_friendly_name(self):
        args = parse_args(
            ["proj", "set-friendly-name", "FRIENDLY_NAME", "--proj=" + self.proj]
        )
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)



import sys
from tq42.exceptions import (
    NoDefaultError,
    InvalidArgumentError,
    NoMatchingAttributeError,
)
from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.client import TQ42Client


def main():
    try:
        args = parse_args(sys.argv[1:])
        env_config_file = args.config
        with TQ42Client(env_config_file) as client:
            res = tq42_all(client, args)
            if res is not None:
                print(res)

    # If you want to catch specific CLI related exceptions
    # you can catch them here and specify custom CLI exception handling
    except NoDefaultError:
        raise NoDefaultError("tq42 {}".format(" ".join(sys.argv[1:]))) from None
    except NoMatchingAttributeError as e:
        raise InvalidArgumentError(
            "tq42 {}".format(" ".join(sys.argv[1:])), e.details
        ) from None

    except Exception as e:
        print(f"Exception: {e}", file=sys.stderr)



import os

import tq42

CONFIG_FOLDER_NAME = ".tq42sdk"


def full_path(basepath, fn=None):
    basepath = os.path.abspath(basepath)
    if fn is None:
        return basepath
    else:
        return os.path.join(basepath, fn)


def code(fn=None):
    basepath = os.path.dirname(tq42.__file__)
    return full_path(basepath, fn=fn)


def tests(fn=None):
    return full_path(code("tests"), fn=fn)


def testdata(fn=None):
    return full_path(tests("testdata"), fn=fn)


def text_files_dir(fn=None):
    return full_path(code("utils/text_files"), fn=fn)


def test_data_dir(fn=None):
    return full_path(code("tests/testdata"), fn=fn)


def cache_dir(fn=None):
    return text_files_dir(fn=fn)


def cache_file():
    return cache_dir("__tq42__.dat")


def config(fn=None):
    return full_path(create_or_get_config_dir(), fn=fn)


def create_or_get_config_dir(alt_config_folder=None):
    package_dir = (
        alt_config_folder if alt_config_folder is not None else get_config_folder_path()
    )

    if not os.path.exists(package_dir):
        os.makedirs(package_dir)

    return package_dir


def get_config_folder_path():
    home_dir = os.path.expanduser("~")
    package_dir = os.path.join(home_dir, CONFIG_FOLDER_NAME)
    return package_dir



import ast

import tq42.cli.cli_functions as cli
from tq42.cli.parsers.params_checker import check_params
from tq42.experiment_run import ExperimentRun
from tq42.algorithm import AlgorithmProto
from tq42.compute import HardwareProto


def experiment_run_group(client, args):
    if args.subcommand == "cancel":
        return exp_run_cancel(client, args)

    elif args.subcommand == "create":
        return exp_run_create(client, args)

    elif args.subcommand == "check":
        return exp_run_check(client, args)

    elif args.subcommand == "list":
        return exp_run_list(client, args)

    elif args.subcommand == "poll":
        return exp_run_poll(client, args)


def exp_run_cancel(client, args):
    res = cli.cancel_exprun(client, args.run)

    return res


def exp_run_poll(client, args):
    res = cli.poll_exprun(client, args.run)

    return res


def exp_run_create(client, args):
    check_params("exp run create", args)
    algo = AlgorithmProto.Value(args.algorithm.upper())
    compute = HardwareProto.Value(args.compute.upper())
    params = ast.literal_eval(args.parameters)
    return ExperimentRun.create(
        client=client,
        algorithm=algo,
        experiment_id=args.exp,
        compute=compute,
        parameters=params,
    ).data


def exp_run_check(client, args):
    return cli.get_exprun(client, args.run)


def exp_run_list(client, args):
    check_params("exp run list", args)
    return cli.list_expruns(client, args.exp)



import unittest

from tq42.functional_tests.functional_test_config import FunctionalTestConfig


class TestTestConfig(unittest.TestCase, FunctionalTestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_functional_test_config(self):
        self.assertIsNotNone(self.org)
        self.assertIsNotNone(self.proj)
        self.assertIsNotNone(self.exp)
        self.assertIsNotNone(self.exp_run)



import pytest

from tq42.dataset import Dataset, StorageProto


def test__get_file_name_from_signed_url():
    signed_url = "https://storage.googleapis.com/random-uuid-that-is-not-valid/photovoltaic_data_units_no_label.xlsx?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=not-valid-cred%40service-account-uri&X-Goog-Date=20240619T071407Z&X-Goog-Expires=900&X-Goog-SignedHeaders=host&X-Goog-Signature=some-random-signature-thingy-with-random-numbers-123913-1240235"
    file_name = Dataset._get_file_name_from_signed_url(signed_url)
    assert file_name == "photovoltaic_data_units_no_label.xlsx"


def test__get_file_name_from_signed_url_no_params():
    signed_url = "https://storage.googleapis.com/random-uuid-that-is-not-valid/photovoltaic_data_units_no_label.xlsx"
    file_name = Dataset._get_file_name_from_signed_url(signed_url)
    assert file_name == "photovoltaic_data_units_no_label.xlsx"


def test__get_file_name_from_signed_url_no_file_ending():
    signed_url = "https://storage.googleapis.com/random-uuid-that-is-not-valid/photovoltaic_data_units_no_label?X-Goog-Algorithm=GOOG4-RSA-SHA256&X-Goog-Credential=not-valid-cred%40service-account-uri&X-Goog-Date=20240619T071407Z&X-Goog-Expires=900&X-Goog-SignedHeaders=host&X-Goog-Signature=some-random-signature-thingy-with-random-numbers-123913-1240235"
    file_name = Dataset._get_file_name_from_signed_url(signed_url)
    assert file_name == "photovoltaic_data_units_no_label"


def test__get_file_name_from_signed_url_no_valid_url():
    signed_url = "hps://storage/random-uuid-that-is-not-valid/photo"

    with pytest.raises(ValueError):
        Dataset._get_file_name_from_signed_url(signed_url)


def test_dataset_export_non_valid_directory():
    data = StorageProto(
        id="83976d81-ffaa-4778-b894-89ca8d5801f2",
        name="pseudo-dataset-data",
    )
    dataset = Dataset.from_proto(client={}, msg=data)
    with pytest.raises(ValueError):
        dataset.export(directory_path="/Users/definitely-not-a-valid-user/directory")



import time
import unittest

from tq42.cli.parsers.tq42parser import parse_args
from tq42.cli.tq42_all import tq42_all
from tq42.experiment_run import ExperimentRun
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig

from tq42.algorithm import AlgorithmProto
from tq42.compute import HardwareProto

from tq42.exceptions import ExperimentRunCancelError


class TestFunctionalInvokeCliExperimentRunGroup(
    unittest.TestCase, FunctionalCLITestConfig
):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_exp_run_list_error(self):
        args = parse_args(["exp", "run", "list"])
        self.assertRaises(SystemExit, tq42_all, self.client, args)

    def test_exp_run_list(self):
        args = parse_args(["exp", "run", "list", "--exp=" + self.exp])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_exp_run_create(self):
        parameters = '{"parameters": {"n": 2, "r": 1, "msg": "correct"}, "inputs": {}}'
        args = parse_args(
            [
                "exp",
                "run",
                "create",
                "--exp=" + self.exp,
                "--compute=SMALL",
                "--algorithm=TOY",
                "--parameters=" + parameters,
            ]
        )
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_exp_run_cancel_too_late(self):
        args = parse_args(["exp", "run", "cancel", self.exp_run])
        self.assertRaises(ExperimentRunCancelError, tq42_all, self.client, args)

    def test_exp_run_cancel(self):
        parameters = {"parameters": {"n": 2, "r": 1, "msg": "correct"}, "inputs": {}}
        exp_run = ExperimentRun.create(
            client=self.client,
            algorithm=AlgorithmProto.TOY,
            experiment_id=self.exp,
            compute=HardwareProto.SMALL,
            parameters=parameters,
        )

        # we need to wait for a small bit to make sure the experiment run can be retrieved from the system
        time.sleep(5)

        args = parse_args(["exp", "run", "cancel", exp_run.id])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)

    def test_exp_run_check(self):
        args = parse_args(["exp", "run", "check", self.exp_run])
        result = tq42_all(self.client, args)
        self.assertIsNotNone(result)



from __future__ import annotations

from typing import Optional, List

from google.protobuf.field_mask_pb2 import FieldMask

from tq42.client import TQ42Client
from tq42.exception_handling import handle_generic_sdk_errors

from com.terraquantum.experiment.v3alpha1.experiment.experiment_pb2 import (
    ExperimentProto,
)
from com.terraquantum.experiment.v3alpha1.experiment.get_experiment_pb2 import (
    GetExperimentRequest,
)
from com.terraquantum.experiment.v3alpha1.experiment.list_experiments_pb2 import (
    ListExperimentsRequest,
)
from com.terraquantum.experiment.v3alpha1.experiment.list_experiments_pb2 import (
    ListExperimentsResponse,
)
from com.terraquantum.experiment.v3alpha1.experiment.update_experiment_request_pb2 import (
    UpdateExperimentRequest,
)

from tq42.utils.utils_for_cache import get_current_value


class Experiment:
    """
    Class to manage experiments
    """

    _client: TQ42Client
    id: str
    data: ExperimentProto

    def __init__(
        self, client: TQ42Client, id: str, data: Optional[ExperimentProto] = None
    ):
        """
        Instantiates an experiment.
        If data is omitted the id is used to get the corresponding data from the API (used like a GET on this id)
        """
        self._client = client
        self.id = id

        if data:
            self.data = data
        else:
            self.data = self._get_data()

    def __repr__(self) -> str:
        return f"<Experiment Id={self.id} Name={self.data.name}>"

    def __str__(self) -> str:
        return str(self.data)

    @handle_generic_sdk_errors
    def _get_data(self) -> ExperimentProto:
        """
        Gets the data corresponding to this experiment id.
        """
        get_exp_request = GetExperimentRequest(id=self.id)
        res = self._client.experiment_client.GetExperiment(
            request=get_exp_request, metadata=self._client.metadata
        )
        return res

    @staticmethod
    def from_proto(client: TQ42Client, msg: ExperimentProto) -> Experiment:
        """
        Creates Experiment instance from a protobuf message.
        """
        return Experiment(client=client, id=msg.id, data=msg)

    @handle_generic_sdk_errors
    def update(self, name: str) -> Experiment:
        """
        Updates this specific experiment
        """
        # Create a new FieldMask instance
        field_mask = FieldMask()

        # Add paths to the FieldMask
        field_mask.paths.append("id")
        field_mask.paths.append("name")

        update_proj_request = UpdateExperimentRequest(
            update_mask=field_mask, id=self.id, name=name
        )
        self.data = self._client.experiment_client.UpdateExperiment(
            request=update_proj_request, metadata=self._client.metadata
        )

        return self

    @handle_generic_sdk_errors
    def set_friendly_name(self, friendly_name: str) -> Experiment:
        """
        Set a friendly name for an experiment.

        For details, see https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#setting-friendly-names-for-projects-and-experiments
        """
        return self.update(name=friendly_name)


@handle_generic_sdk_errors
def list_all(client: TQ42Client, project_id: Optional[str] = None) -> List[Experiment]:
    """
    List all the experiments you have permission to view within a specific project.
    If no project_id is specified the currently set project id will be used for this.

    For details, see https://terra-quantum-tq42sdk-docs.readthedocs-hosted.com/en/latest/Python_Developer_Guide/Setting_Up_Your_Environment.html#list-all-experiments
    """
    if not project_id:
        project_id = get_current_value("proj")

    list_experiments_request = ListExperimentsRequest(project_id=project_id)
    res: ListExperimentsResponse = client.experiment_client.ListExperiments(
        request=list_experiments_request, metadata=client.metadata
    )
    return [
        Experiment.from_proto(client=client, msg=experiment_run)
        for experiment_run in res.experiments
    ]



import textwrap
import json
from argparse import RawTextHelpFormatter

from tq42.utils import dirs


class Help:
    def __init__(self, name, level, text=None, cli_url=None, python_url=None):
        self.name = name
        self.level = level
        self.cli_url = cli_url
        self.python_url = python_url
        self.text = text
        self.argparser = None

    def add_parser(self, subparser):
        self.argparser = subparser.add_parser(
            self.name,
            description=self.display_help,
            help=self.text,
            formatter_class=RawTextHelpFormatter,
        )
        return self.argparser

    def add_flag(self, flag, help_flags=None):
        if help_flags is None:
            help_flags = default_help_flags

        help_flag = help_flags[flag]
        help_flag.add_argument(self.argparser)

    @property
    def display_help(self):
        text = self.text
        if self.cli_url is not None:
            text = f"{textwrap.dedent(text)}\n\nFor details, see {self.cli_url}"
        return text

    @property
    def display_docstring(self):
        text = self.text
        if self.python_url is not None:
            text = f"{textwrap.dedent(text)}\n\nFor details, see {self.python_url}"
        return text

    @staticmethod
    def from_dict(d: dict):
        return Help(
            name=d["name"],
            level=d["level"],
            text=d.get("text", None),
            cli_url=d.get("cli_url", None),
            python_url=d.get("python_url", None),
        )

    @staticmethod
    def walk_tree(nodes, tree: dict = None):
        if tree is None:
            tree = default_help_tree
        d = tree
        for node in nodes:
            d = d[node]
        d = d["help"].copy()
        if len(nodes) > 0:
            name = nodes[-1]
        else:
            name = "tq42"
        d["name"] = name
        return d

    @staticmethod
    def lookup_help(nodes, tree: dict = None):
        d = Help.walk_tree(nodes, tree)
        return Help.from_dict(d)


def get_help_tree():
    with open(dirs.text_files_dir("help.json")) as f:
        return json.load(f)


default_help_tree = get_help_tree()


def get_commands():
    global_help_object = Help.lookup_help([])
    cli_url = global_help_object.cli_url
    tq42_commands_file_path = dirs.full_path(dirs.text_files_dir(), "tq42_commands.txt")
    with open(tq42_commands_file_path, "r") as f:
        return str(f.read()).format(cli_url)


class HelpFlag:
    def __init__(self, flag, text, metavar=None, short_flag=None):
        self.flag = flag
        self.text = text
        self.metavar = metavar
        self.short_flag = short_flag

    @staticmethod
    def from_dict(d: dict):
        return HelpFlag(
            flag=d["flag"],
            text=d["text"],
            metavar=d.get("metavar", None),
            short_flag=d.get("short_flag", None),
        )

    @staticmethod
    def flag_list_to_dict(a):
        return {d["flag"].replace("--", ""): HelpFlag.from_dict(d) for d in a}

    def add_argument(self, parser):
        if self.short_flag is None:
            parser.add_argument(self.flag, help=self.text, metavar=self.metavar)
        else:
            parser.add_argument(
                self.short_flag, self.flag, help=self.text, metavar=self.metavar
            )


def get_flags():
    with open(dirs.text_files_dir("help_flags.json")) as f:
        return HelpFlag.flag_list_to_dict(json.load(f))


default_help_flags = get_flags()



import argparse
import unittest

import tq42.cli.cli_functions as cli
from tq42.functional_tests.functional_test_config import FunctionalCLITestConfig


class TestFunctionalCli(unittest.TestCase, FunctionalCLITestConfig):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_proj_set(self):
        result = cli.set_project(self.client, self.proj)
        self.assertIsNotNone(result)

    def test_proj_show(self):
        result = cli.proj_show(self.client)
        success = "org=" in str(result)
        self.assertEqual(True, success)

    def test_list_orgs(self):
        result = cli.list_orgs(self.client)
        success = "org=" in str(result)
        self.assertEqual(True, success)

    def test_list_proj_by_org(self):
        result = cli.list_proj_by_org(self.client, self.org)
        success = "proj=" in str(result)
        self.assertEqual(True, success)

    def test_list_projects(self):
        parser = argparse.ArgumentParser()

        parser.add_argument(
            "--org", type=str, help="Ex: 1df1f520-1834-4d3e-ac1d-65b440fd8f3c"
        )
        args = parser.parse_args(["--org", self.org])
        result = cli.list_proj_by_org(self.client, args.org)
        success = "proj=" in str(result)
        self.assertEqual(True, success)

    def test_get_exprun(self):
        result = cli.get_exprun(self.client, self.exp_run)
        success = "status" in str(result)
        self.assertEqual(True, success)



import os.path
import time
import unittest

from tq42.dataset import Dataset, list_all as list_all_datasets, DatasetSensitivityProto
from tq42.functional_tests.functional_test_config import FunctionalTestConfig

from com.terraquantum.storage.v1alpha1.storage_pb2 import StorageStatusProto


class TestFunctionalTQ42APIDataset(unittest.TestCase, FunctionalTestConfig):
    dataset_id: str
    dataset_name = "Dataset Name Test1"

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_proj_dataset_create(self):
        dataset = Dataset.create(
            client=self.get_client(),
            project_id=self.proj,
            name=self.dataset_name,
            description="Dataset description test 1",
            url="gs://circuit-runner-demo",
            sensitivity=DatasetSensitivityProto.CONFIDENTIAL,
        )
        TestFunctionalTQ42APIDataset.dataset_id = dataset.id
        self.assertEqual("Dataset Name Test1", dataset.data.name)

        # make sure dataset is properly available
        time.sleep(3)

        # waiting for the dataset transfer to be completed
        while dataset.data.status not in [
            StorageStatusProto.FAILED,
            StorageStatusProto.COMPLETED,
            StorageStatusProto.EMPTY,
        ]:
            dataset._refresh()
            time.sleep(3)

        self.assertEqual(StorageStatusProto.COMPLETED, dataset.data.status)

    def test_proj_dataset_get(self):
        dataset = Dataset(
            client=self.get_client(), id=TestFunctionalTQ42APIDataset.dataset_id
        )
        self.assertEqual(self.dataset_name, dataset.data.name)

    def test_proj_datasets_list(self):
        datasets = list_all_datasets(client=self.get_client(), project_id=self.proj)
        found_id = any(
            TestFunctionalTQ42APIDataset.dataset_id in dataset.id
            for dataset in datasets
        )
        self.assertTrue(found_id)

    def test_proj_dataset_export(self):
        dataset_export = Dataset(
            client=self.get_client(), id=TestFunctionalTQ42APIDataset.dataset_id
        ).export(".")

        assert 1 == len(dataset_export)
        assert os.path.exists(dataset_export[0])

        try:
            os.remove(dataset_export[0])
        except OSError:
            pass

        assert not os.path.exists(dataset_export[0])
